-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- PicoBlaze compatible UART Transceiver
--
-- Data register is shared by Rx (read) and Tx (write).
-- Status register contains basic UART status information.
-- Once the UART is implemented, its speed cannot be changed.
--
-- Rx interrupt cleared by reading from data register
-- Tx interrupt is only a short pulse
--
-- Status:
--  CODING COMPLETE     YES
--  SIMULATED           YES - OK
--  TESTED              YES
--
---------------------------------------------------------------------------
---------------------------------------------------------------------------
-- Status/control register:
-- [0] Tx IRQ enable
-- [1] Rx IRQ enable
-- [2] Tx ready
-- [3] new Rx data
-- [4] Rx buffer overflow
-- [5] -( unused )-
-- [6] -( unused )-
-- [7] -( unused )-
----------------------------------------------------------------------
----------------------------------------------------------------------
--   period vale for typical data rates and frequences   |-----------|
----------------------------------------------------------------------
-------------|              clk  frequency               |-----------|
---------------------------------------------------------| max cable |
-- Baud Rate |   50 MHz |  100 MHz | 62,5 MHz |19,44 MHz |  length   |
-------------|----------|----------|----------|----------|-----------|
--      110  |  454 545 |  909 091 |  568 182 |  176 727 |           |
--      300  |  166 667 |  333 333 |  208 333 |   64 800 |           |
--      600  |   83 333 |  166 667 |  104 167 |   32 400 |           |
--    1 200  |   41 667 |   83 333 |   52 083 |   16 200 |           |
--    2 400  |   20 833 |   41 667 |   26 042 |    8 100 |   900m    |
--    4 800  |   10 417 |   20 833 |   13 021 |    4 050 |   300m    |
--    9 600  |    5 208 |   10 417 |    6 510 |    2 025 |   150m    |
--   14 400  |    3 472 |    6 944 |    4 340 |    1 350 |           |
--   19 200  |    2 604 |    5 208 |    3 255 |    1 013 |    15m    |
--   38 400  |    1 302 |    2 604 |    1 628 |      506 |           |
--   57 600  |      868 |    1 736 |    1 085 |      338 |           |
--  115 200  |      434 |      868 |      543 |      169 |           |
----------------------------------------------------------------------
---------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
-----------------------------------------------------------------------
ENTITY UART_block IS
    GENERIC(data_port_ID: STD_LOGIC_VECTOR(7 DOWNTO 0):= "00000001";-- UART Tx/Rx data register address
            Stat_port_ID: STD_LOGIC_VECTOR(7 DOWNTO 0):= "00000010";-- Status register address
            period:       INTEGER := 5208);                         -- 9600 Bd  @  50 MHz
    PORT( clk:            IN  STD_LOGIC;                            -- system clock
          Rx:             IN  STD_LOGIC;                            -- UART serial input
          Tx:             OUT STD_LOGIC:= '0';                      -- UART serial output
          Data_in:        IN  STD_LOGIC_VECTOR(7 DOWNTO 0);         -- Data IN (PicoBlaze OUT)
          Data_out:       OUT STD_LOGIC_VECTOR(7 DOWNTO 0):= X"00"; -- Data reg OUT (to PicoBlaze input MUX)
          Stat_out:       OUT STD_LOGIC_VECTOR(7 DOWNTO 0):= X"00"; -- Status reg OUT (to PicoBlaze input MUX)
          Address:        IN  STD_LOGIC_VECTOR(7 DOWNTO 0);         -- PicoBlaze Port ID
          read_strobe:    IN  STD_LOGIC;                            -- PicoBlaze read strobe
          write_strobe:   IN  STD_LOGIC;                            -- PicoBlaze write strobe
          Tx_ready_int:   OUT STD_LOGIC := '0';                     -- Tx done interrupt request
                                                                    -- edge only (short pulse captured by
                                                                    -- interrupt controller)
          Rx_new_int:     OUT STD_LOGIC := '0'                      -- New Rx data interrupt request
                                                                    -- celared by reading the data reg
          );
END UART_block;
-----------------------------------------------------------------------
ARCHITECTURE Behavioral OF UART_block IS
-----------------------------------------------------------------------
  COMPONENT UART_Basic_Tx
    GENERIC( period:   INTEGER := 5208);
    PORT( clk:            IN  STD_LOGIC;                    -- System clock
          Data:           IN  STD_LOGIC_VECTOR(7 DOWNTO 0); -- data to be sent
          Start:          IN  STD_LOGIC;                    -- start of transmision
          Tx:             OUT STD_LOGIC;                    -- UART output
          Ready:          OUT STD_LOGIC := '1');            -- ready to sent next byte
  END COMPONENT;
-----------------------------------------------------------------------
  COMPONENT UART_Basic_Rx
    GENERIC( period:   INTEGER := 5208);
    PORT( clk:            IN  STD_LOGIC;                    -- system clock
          Rx:             IN  STD_LOGIC;                    -- UART serial input
          Data:           OUT STD_LOGIC_VECTOR(7 DOWNTO 0); -- received data
          Data_Ready:     OUT STD_LOGIC);                   -- new data received {interrupt, rising edge!}
  END COMPONENT;
-----------------------------------------------------------------------
  SIGNAL Tx_ready:      STD_LOGIC := '0';
  SIGNAL Tx_start:      STD_LOGIC := '0';
  SIGNAL Tx_data:       STD_LOGIC_VECTOR(7 DOWNTO 0):= (OTHERS => '0');
  SIGNAL Rx_new:        STD_LOGIC := '0';
  SIGNAL Rx_data:       STD_LOGIC_VECTOR(7 DOWNTO 0):= (OTHERS => '0');

-- unused; Instead of tri-state bus, a MUX is used on higher level
--SIGNAL Data_int:      STD_LOGIC_VECTOR(7 DOWNTO 0):= (OTHERS => '0');     -- internal signal; input to tri-state output buffer

  SIGNAL Stat_reg:      STD_LOGIC_VECTOR(7 DOWNTO 0) := X"04";

  SIGNAL Tx_ready_buf:  STD_LOGIC := '0';                                   -- delayed signal for edge detection
-----------------------------------------------------------------------
BEGIN
-----------------------------------------------------------------------
-----------------------------------------------------------------------
  inst_UART_Basic_Tx : UART_Basic_Tx GENERIC MAP(
    period          => period)
  PORT MAP(
    clk             => clk,
    Data            => Tx_data,
    Start           => Tx_start,
    Tx              => Tx,
    Ready           => Tx_ready);
-----------------------------------------------------------------------
-----------------------------------------------------------------------
  inst_UART_Basic_Rx : UART_Basic_Rx GENERIC MAP(
    period          => period)
  PORT MAP(
    clk             => clk,
    Rx              => Rx,
    Data            => Rx_data,
    Data_Ready      => Rx_new);
  ---------------------------------------------------------------------
  Stat_reg(7 DOWNTO 5) <= "000";
-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- Data Input
  IN_conrol: PROCESS(clk) BEGIN
    IF clk'EVENT AND clk = '1' THEN
      -----------------------------------------------------------------
      IF Address = data_port_ID AND write_strobe = '1' THEN
        Tx_data <= Data_in;
      END IF;
      -----------------------------------------------------------------
      IF Address = stat_port_ID AND write_strobe = '1' THEN
        Stat_reg(1 DOWNTO 0) <= Data_in(1 DOWNTO 0);
      END IF;
      -----------------------------------------------------------------
    END IF;
  END PROCESS IN_conrol;
-----------------------------------------------------------------------
-----------------------------------------------------------------------
  Data_out <= Rx_data;
  Stat_out <= Stat_reg;
-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- Data Output
--  OUT_conrol: PROCESS(clk) BEGIN
--    IF clk'EVENT AND clk = '1' THEN
--      -----------------------------------------------------------------
--      IF Address = data_port_ID THEN
--        Data_out <= Rx_data;
--      END IF;
--      -----------------------------------------------------------------
--      IF Address = stat_port_ID THEN
--        Data_out <= Stat_reg;
--      END IF;
--      -----------------------------------------------------------------
--    END IF;
--  END PROCESS OUT_conrol;
-----------------------------------------------------------------------
-----------------------------------------------------------------------

-- CASE statement cause warnings during synthesis (data_port_ID and
-- stat_port_ID aren't constants)
--
--      -----------------------------------------------------------------
--      CASE Address IS
--        ---------------------------------------------------------------
--        -- UART Data port selected
--        WHEN data_port_ID =>
--          IF read_strobe = '1'  THEN Data_out <= Rx_data;
--                                ELSE Data_out <= (OTHERS => 'Z'); END IF;
--          IF write_strobe = '1' THEN Tx_data <= Data_in; END IF;
--        ---------------------------------------------------------------
--        -- UART Status port selected
--        WHEN stat_port_ID =>
--          IF read_strobe = '1'  THEN Data_out <= Stat_reg;
--                                ELSE Data_out <= (OTHERS => 'Z'); END IF;
--          IF write_strobe = '1' THEN Stat_reg(1 DOWNTO 0) <= Data_in(1 DOWNTO 0); END IF;
--        ---------------------------------------------------------------
--        -- Other (non-UART) port selected
--        WHEN OTHERS => Data_out <= (OTHERS => 'Z');
--        ---------------------------------------------------------------
--      END CASE;
--      -----------------------------------------------------------------

--  END PROCESS IO_conrol;
-----------------------------------------------------------------------
-- status register control and interrupts

-- Status register:
-- [0] Tx interrupt enable
-- [1] Rx interrupt enable
-- [2] Tx ready
-- [3] new Rx data
-- [4] Rx buffer overflow

-- Tx_ready_int is activated on rising edge of Tx_ready signal
-- Rx_new_int is activated when Rx_new signal is asserted

  int_ctrl: PROCESS(clk) BEGIN
    IF clk'EVENT AND clk = '1' THEN
      -----------------------------------------------------------------
      Tx_ready_buf <= Tx_ready;     -- pipelining for edge detection
      -----------------------------------------------------------------
      -- Tx Interrupt
      -- rising edge on Tx_ready and Tx_interrupt is enabled
      IF Tx_ready_buf = '0' AND Tx_ready = '1' AND Stat_reg(0) = '1' THEN
        Tx_ready_int <= '1';
      ELSE
        Tx_ready_int <= '0';
      END IF;

      -----------------------------------------------------------------
      -- Rx Interrupt
      -- Rx_new is active and Rx_interrupt is enabled
      IF Rx_new = '1' AND Stat_reg(1) = '1' THEN
        Rx_new_int <= '1';
      ELSE
        Rx_new_int <= '0';
      END IF;

      -----------------------------------------------------------------
      -- "New Rx data" flag and "Rx buffer overflow" flag
      -- It is more important to set flags than to reset them
      IF Rx_new = '1' THEN
        stat_reg(3) <= '1';             -- new data ready FLAG is set
        IF stat_reg(3) = '1' THEN       -- Rx buffer already full
          stat_reg(4) <= '1';           -- overflow FLAG is set
        END IF;
      ELSE
        ---------------------------------------------------------------
        -- when reading from data register
        IF Address = data_port_ID AND read_strobe = '1' THEN
          stat_reg(3) <= '0';     -- clear flags
        END IF;
        ---------------------------------------------------------------
        -- when reading from status register
        IF Address = stat_port_ID AND read_strobe = '1' THEN
          stat_reg(4) <= '0';     -- clear overflow flag when reading status register
        END IF;
      END IF;

      -----------------------------------------------------------------
      -- "Tx ready" flag
      stat_reg(2) <= Tx_ready;

      -----------------------------------------------------------------
      -- UART Tx start control - start Tx when new data are written to data register
      IF Address = data_port_ID AND write_strobe = '1' THEN
        Tx_start <= '1';
      ELSE
        Tx_start <= '0';
      END IF;

      -----------------------------------------------------------------
    END IF;
  END PROCESS int_ctrl;
-----------------------------------------------------------------------
-----------------------------------------------------------------------
END Behavioral;
