//========================================================================
//  CFSOCHANNEL.CC - part of
//
//                 OMNeT++/OMNEST
//              Discrete System Simulation in C++
//
//========================================================================

/*--------------------------------------------------------------*
  Copyright (C) 1992-2008 Andras Varga
  Copyright (C) 2006-2008 OpenSim Ltd.

  This file is distributed WITHOUT ANY WARRANTY. See the file
  `license' for details on this and other legal matters.
*--------------------------------------------------------------*/

#include "cfsochannel.h"
#include "cmessage.h"
#include "cmodule.h"
#include "cenvir.h"
#include "random.h"
#include "distrib.h"
#include "csimulation.h"
#include "globals.h"
#include "cgate.h"
#include "cexception.h"
#include "ctimestampedvalue.h"

#ifdef WITH_PARSIM
#include "ccommbuffer.h"
#endif

//USING_NAMESPACE

using std::ostream;

Register_Class(cFSOChannel);
Define_Channel(cFSOChannel);


simsignal_t cFSOChannel::channelBusySignal;
simsignal_t cFSOChannel::messageSentSignal;
simsignal_t cFSOChannel::messageDiscardedSignal;

cFSOChannel::cFSOChannel(const char *name) : cChannel(name)
{
    txfinishtime = 0;
    delay = 0;
    datarate = 100000000;
}

cFSOChannel::~cFSOChannel()
{
}

cFSOChannel *cFSOChannel::create(const char *name)
{
    return dynamic_cast<cFSOChannel *>(cChannelType::getDatarateChannelType()->create(name));
}

std::string cFSOChannel::info() const
{
    return cChannel::info();
}

void cFSOChannel::initialize()
{
    channelBusySignal = registerSignal("channelBusy");
    messageSentSignal = registerSignal("messageSent");
    messageDiscardedSignal = registerSignal("messageDiscarded");

    emit(channelBusySignal, 0);
}

void cFSOChannel::rereadPars()
{
    delay = par("delay");
    datarate = par("datarate");

    setFlag(FL_ISDISABLED, par("disabled"));
    setFlag(FL_DELAY_NONZERO, delay!=0);
    setFlag(FL_DATARATE_NONZERO, datarate!=0);
}

void cFSOChannel::handleParameterChange(const char *)
{
    rereadPars();
}

void cFSOChannel::setDelay(double d)
{
    par("delay").setDoubleValue(d);
}

void cFSOChannel::setDatarate(double d)
{
    par("datarate").setDoubleValue(d);
}

void cFSOChannel::setDisabled(bool d)
{
    par("disabled").setBoolValue(d);
}

simtime_t cFSOChannel::calculateDuration(cMessage *msg) const
{
    if (flags & FL_DATARATE_NONZERO && msg->isPacket())
        return ((cPacket *)msg)->getBitLength() / datarate;
    else
        return SIMTIME_ZERO;
}

void cFSOChannel::processMessage(cMessage *msg, simtime_t t, result_t& result)
{
    if (flags & FL_ISDISABLED)
    {
        result.discard = true;
        cTimestampedValue tmp(t, msg);
        emit(messageDiscardedSignal, &tmp);
        return;
    }

    // datarate modeling
    if ((flags & FL_DATARATE_NONZERO) && msg->isPacket())
    {
        cPacket *pkt = (cPacket *)msg;
        simtime_t duration = pkt->getBitLength() / datarate;
        result.duration = duration;
        txfinishtime = t + duration;
    }
    else {
        txfinishtime = t;
    }

    // propagation delay modeling
    result.delay = delay;

    // bit error modeling

    float_t ber = 0.1;

    if ((ber > 0.0) && msg->isPacket())
    {
        cPacket *pkt = (cPacket *)msg;
        if (dblrand() < 1.0 - pow(1.0-ber, (double)pkt->getBitLength()))
            pkt->setBitError(true);
    }

    // emit signals
    if (mayHaveListeners(messageSentSignal))
    {
        MessageSentSignalValue tmp(t, msg, &result);
        emit(messageSentSignal, &tmp);
    }
    if (mayHaveListeners(channelBusySignal))
    {
        cTimestampedValue tmp(t, 1L);
        emit(channelBusySignal, &tmp);

        tmp.timestamp = txfinishtime; tmp.l = 0L;
        emit(channelBusySignal, &tmp);
    }
}

void cFSOChannel::forceTransmissionFinishTime(simtime_t t)
{
    //TODO record this into the eventlog so that it can be visualized in the sequence chart
    txfinishtime = t;
}
